//=================================================================================
//
//=================================================================================

// varBlock stores an array of arbitrary length varGeom objects into
// a variable length array of size baseType. The varGeom object must have
// the following methods:
/*
template <typename coordinate_T> class varGeom {
public:
	// Accessors
	int recLen() const;	// length of object in bytes.
	bool isRemoved() const;	// object has been removed from the database.
	varBox<coordinate_T> bbox() const;	// bounding box of object
	
	// Manipulators
	void setRemoved();
};

When iterating over the objects in the block the objects will be returned in
xy sort order x first y second of bbox().lowerLeft. Any object that isRemoved()
will be either skipped or removed entirely.
*/

// class varShapeBlock is a variable array that keeps track of where the information
// is stored whether on disk or in RAM

template <typename baseType,typename coordinate_T,typename varGeom> class varBlock {
public:
	varBlock(): _count(0),_end(0),_used(0),_array(0) {}
	~varBlock() { delete [] _array; }

	//==============================================
	// Accessors
	//==============================================
	bool empty() { return _end == 0; }
	uint8 count() { return _count; }
	uint16 used() { return _used; }
	varGeom* operator[uint16 location] {
#ifdef GDB_SANITY
		if (location < 0 || location > _used) {
		    return 0;
		}
#endif
		return static_cast<varGeom<coordinate_T>*>&_array[location];
	}
    Box dataBox() const { return _dataBox; }

	//==============================================
	// Manipulators
	//==============================================
	// Allocate space for an object in the array in bytes
	// When _array needs to grow it grows by a multiple of delta.
	// If delta == 0 then array doubles in size when it needs to grow
	void* allocate(uint32 numBytes);
	void free(varGeom* obj);
	// remove all objects from the array.
	void clear();

	//==============================================
    // Static methods
	//==============================================
	static void setDelta(uint16 v) { _delta = v; }
	static uint16 getDelta() { return _delta; }

private:
    inline uint32 baseTypeLen(uint32 byteSize)
	{
		return (byteSize + sizeof(baseType) - 1)/sizeof(baseType);
	}
	void grow(uint32 numBaseType,uint16 delta);

// Common to both array and block
	Box			_dataBox;	// minimum Box containing all varGeom objects
	baseType*	_array;		// storage
#ifdef GDB_DISKIO
	off_t		_diskLocation;	// location of data stored on disk
#endif
	uint16		_used;		// number of used baseType entries
	uint16		_end;		// Length of array in baseType;
	uint8		_count;		// Number of varGeom objects in block
	bool		_sorted;	// returns true if the objects in the block are sorted.
	bool		_boxValid;	// returns true of dataBox is valid.
#ifdef GDB_DISKIO
	bool		_modified;	// returns true if data in memory has changed
	vFileId		_file;
#endif		
};

template <typename baseType,typename varGeom>
bool
varBlock<baseType,varGeom>::grow(uint32 numBaseType,uint16 delta)
{
    uint32 growSize = numBaseType - (_end - _used);
	if (delta == 0) {
	    if (growSize < _end) {
			growSize = _end;
		}
	} else {
		growSize = delta * ((growSize + delta - 1)/delta);
	}
	if (_end + growSize > maxSignedValue(bitSizeof(_end))) {
	    return false;
	}
	baseType* newArray = new baseType[_end + growSize];
	memcpy(newArray,_array,sizeof(baseType * _used));
	delete [] _array;
	_array = newArray;
	_end += growSize;
	return true;
}

template <typename baseType,typename varGeom>
void*
varBlock<baseType,varGeom>::allocate(uint32 numBytes,uint16 delta)
{
    if (_count + 1 > MAXLEN) {
		return 0;
	}
	int numBaseType = baseTypeLen(numBytes);
	if (_used + numBaseType > _end) {
	    if (!grow(numBaseType)) {
		    return 0;
		}
	}
	void* ptr = &_array[_used];
	_used += numBaseType;
	if (byteReclen != sizeof(baseType) * numBaseType) {
		// Pad end with 0 to avoid purify errors.
		int padLen = byteReclen - (sizeof(baseType) * numBaseType);
		char* start = static_cast<char*>(&_array[used]) - padLen;
		memset(start,0,padlen);
	}
	++_count;
	return ptr;
}

template <typename baseType,typename varGeom>
void
varBlock<baseType,varGeom>::free(varGeom* obj)
{
    if (!obj) {
	    return;
	}
	int32 location = static_cast<baseType*>(obj) - _array;
	if (location < 0 || location > _used) {
	    cerr << "###ERROR: free nonexistant object." << endl;
		return;
	}
	int32 numBaseType = baseTypeLen(obj->recLen());
	int32 objEnd = location + numBaseType;
	if (objEnd > _used) {
	    cerr << "###ERROR: free misaligned object." << endl;
		return;
	}
	if (objEnd != _used) {
	    memcpy(&_array[location],&_array[objEnd],
			(_used - objEnd)*sizeof(baseType));

	}
	_used -= numBaseType;
	--_count;
}

template <typename baseType,typename varGeom>
void
varBlock<baseType,varGeom>::clear()
{
    if (_array) {
	    delete [] _array;
		_array = 0;
		_end = 0;
	}
}


