//===========================================================================
//  shape.h  -  geometric object information
//  -----------------------------------------------
//  begin                : Thu Jun 3 2004
//  copyright            : (C) 2004 by Stan Benedict
//  email                : stanb@cybrquest.com
//===========================================================================


#ifndef SHAPE_H
#define SHAPE_H

#include "types.h"

//============================================================================
// Define geometric objects as templates based on coordinate_T so that the
// objects can be configured to consume the minimum space necessary to store
// geometric information by applying an external offset to the geometric values.
// The var prefix indicates that these objects are variable based on coordinate_T
//============================================================================

#define coordType template <typename coordinate_T>
#define XY varXY<coordinate_T>
#define Box varBox<coordinate_T>
#define Shape varShape<coordinate_T>

coordType class varXY {
public:
	varXY(): _x(0),_y(0) {}
	varXY(coordinate_T v): _x(v),_y(v) {}
	varXY(coordinate_T x,coordinate_T y): _x(x),_y(y) {}

	//============================================================
	// Accessors
	//============================================================
	coordinate_T x() const { return _x; }
	coordinate_T y() const { return _y; }
	
	//============================================================
	// Manipulators
	//============================================================
	void x(coordinate_T v) { _x = v; }
	void y(coordinate_T v) { _y = v; }
	
private:
	coordinate_T _x;
	coordinate_T _y;
};

coordType class varBox {
public:
    varBox():
		_lowerLeft(XY(maxSignedValue(sizeof(coordinate_T)))),
		_upperRight(XY(minSignedValue(sizeof(coordinate_T))))
	{}
    varBox(XY& lowerLeft,XY& upperRight):
	    _lowerLeft(lowerLeft),_upperRight(upperRight) {}
	~varBox() {}
	
	//=================================================================
	// Accessors
	//=================================================================
	XY& lowerLeft() { return _lowerLeft; }
	XY& upperRight() { return _upperRight; }
	coordinate_T xDimension() { return _upperRight.x() - _lowerLeft.x(); }
	coordinate_T yDimension() { return _upperRight.y() - _lowerLeft.y(); }
	
	bool operator<(const Box& box) {
		return (lowerLeft().x() < box.lowerLeft().x()) ? true :
			lowerLeft().x() == box.lowerLeft().x() &&
				lowerLeft().y() < box.lowerLeft().y();
	}
	
	//=================================================================
	// Manipulators
	//=================================================================
	void operator|=(const Box& box)
	{
	    if (lowerLeft().x() > box.lowerLeft().x()) {
			_lowerLeft.x(box.lowerLeft().x());
		}
		if (lowerLeft().y() > box.lowerLeft().y()) {
			_lowerLeft.y(box.lowerLeft().y());
		}
		if (upperRight().x() < box.upperRight().x()) {
			_upperRight.x(box.upperRight().x());
		}
		if (upperRight().y() < box.upperRight().y()) {
			_upperRight.y(box.upperRight().y());
		}
	}
	
	Box& operator=(const Box& box)
	{
		_lowerLeft = box.lowerLeft();
		_upperRight = box.upperRight();
		return *this;
	}
private:
	XY _lowerLeft;
	XY _upperRight;
};



coordType class varShape {
public:
  varShape() {}
	varShape(Box& box):
    _box(box) {}
	~varShape() {}

  Box box() { return _box; }
//   int32 id() { return -1; }  Is this needed?
  static int16 reclen() { sizeof(varShape); }
private:
	Box _box;
};


// This is a shape which stores a handle id (reference)
coordType class varRefShape: public varShape<coordinate_T> {
public:
	varRefShape(): _id(-1) {}
	varRefShape(Box& box,int32 id): varShape<coordinate_T>(box),_id(id) {}
	~varRefShape() {}

  int32 id() { return _id; }
  int16 reclen() { return sizeof(varRefShape); }
private:
	int32 _id;
};


#endif

